# Copyright 2011 Canonical Ltd.
#
# This file is part of u1db.
#
# u1db is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# u1db is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with u1db.  If not, see <http://www.gnu.org/licenses/>.

"""The in-memory Database class for U1DB."""

import json

from leap.soledad.common.l2db import (
    Document, errors,
    query_parser, vectorclock)
from leap.soledad.common.l2db.backends import CommonBackend, CommonSyncTarget


def get_prefix(value):
    key_prefix = '\x01'.join(value)
    return key_prefix.rstrip('*')


class InMemoryDatabase(CommonBackend):
    """A database that only stores the data internally."""

    def __init__(self, replica_uid, document_factory=None):
        self._transaction_log = []
        self._docs = {}
        # Map from doc_id => [(doc_rev, doc)] conflicts beyond 'winner'
        self._conflicts = {}
        self._other_generations = {}
        self._indexes = {}
        self._replica_uid = replica_uid
        self._factory = document_factory or Document

    def _set_replica_uid(self, replica_uid):
        """Force the replica_uid to be set."""
        self._replica_uid = replica_uid

    def set_document_factory(self, factory):
        self._factory = factory

    def close(self):
        # This is a no-op, We don't want to free the data because one client
        # may be closing it, while another wants to inspect the results.
        pass

    def _get_replica_gen_and_trans_id(self, other_replica_uid):
        return self._other_generations.get(other_replica_uid, (0, ''))

    def _set_replica_gen_and_trans_id(self, other_replica_uid,
                                      other_generation, other_transaction_id):
        self._do_set_replica_gen_and_trans_id(
            other_replica_uid, other_generation, other_transaction_id)

    def _do_set_replica_gen_and_trans_id(self, other_replica_uid,
                                         other_generation,
                                         other_transaction_id):
        # TODO: to handle race conditions, we may want to check if the current
        #       value is greater than this new value.
        self._other_generations[other_replica_uid] = (other_generation,
                                                      other_transaction_id)

    def get_sync_target(self):
        return InMemorySyncTarget(self)

    def _get_transaction_log(self):
        # snapshot!
        return self._transaction_log[:]

    def _get_generation(self):
        return len(self._transaction_log)

    def _get_generation_info(self):
        if not self._transaction_log:
            return 0, ''
        return len(self._transaction_log), self._transaction_log[-1][1]

    def _get_trans_id_for_gen(self, generation):
        if generation == 0:
            return ''
        if generation > len(self._transaction_log):
            raise errors.InvalidGeneration
        return self._transaction_log[generation - 1][1]

    def put_doc(self, doc):
        if doc.doc_id is None:
            raise errors.InvalidDocId()
        self._check_doc_id(doc.doc_id)
        self._check_doc_size(doc)
        old_doc = self._get_doc(doc.doc_id, check_for_conflicts=True)
        if old_doc and old_doc.has_conflicts:
            raise errors.ConflictedDoc()
        if old_doc and doc.rev is None and old_doc.is_tombstone():
            new_rev = self._allocate_doc_rev(old_doc.rev)
        else:
            if old_doc is not None:
                if old_doc.rev != doc.rev:
                    raise errors.RevisionConflict()
            else:
                if doc.rev is not None:
                    raise errors.RevisionConflict()
            new_rev = self._allocate_doc_rev(doc.rev)
        doc.rev = new_rev
        self._put_and_update_indexes(old_doc, doc)
        return new_rev

    def _put_and_update_indexes(self, old_doc, doc):
        for index in self._indexes.itervalues():
            if old_doc is not None and not old_doc.is_tombstone():
                index.remove_json(old_doc.doc_id, old_doc.get_json())
            if not doc.is_tombstone():
                index.add_json(doc.doc_id, doc.get_json())
        trans_id = self._allocate_transaction_id()
        self._docs[doc.doc_id] = (doc.rev, doc.get_json())
        self._transaction_log.append((doc.doc_id, trans_id))

    def _get_doc(self, doc_id, check_for_conflicts=False):
        try:
            doc_rev, content = self._docs[doc_id]
        except KeyError:
            return None
        doc = self._factory(doc_id, doc_rev, content)
        if check_for_conflicts:
            doc.has_conflicts = (doc.doc_id in self._conflicts)
        return doc

    def _has_conflicts(self, doc_id):
        return doc_id in self._conflicts

    def get_doc(self, doc_id, include_deleted=False):
        doc = self._get_doc(doc_id, check_for_conflicts=True)
        if doc is None:
            return None
        if doc.is_tombstone() and not include_deleted:
            return None
        return doc

    def get_all_docs(self, include_deleted=False):
        """Return all documents in the database."""
        generation = self._get_generation()
        results = []
        for doc_id, (doc_rev, content) in self._docs.items():
            if content is None and not include_deleted:
                continue
            doc = self._factory(doc_id, doc_rev, content)
            doc.has_conflicts = self._has_conflicts(doc_id)
            results.append(doc)
        return (generation, results)

    def get_doc_conflicts(self, doc_id):
        if doc_id not in self._conflicts:
            return []
        result = [self._get_doc(doc_id)]
        result[0].has_conflicts = True
        result.extend([self._factory(doc_id, rev, content)
                       for rev, content in self._conflicts[doc_id]])
        return result

    def _replace_conflicts(self, doc, conflicts):
        if not conflicts:
            del self._conflicts[doc.doc_id]
        else:
            self._conflicts[doc.doc_id] = conflicts
        doc.has_conflicts = bool(conflicts)

    def _prune_conflicts(self, doc, doc_vcr):
        if self._has_conflicts(doc.doc_id):
            autoresolved = False
            remaining_conflicts = []
            cur_conflicts = self._conflicts[doc.doc_id]
            for c_rev, c_doc in cur_conflicts:
                c_vcr = vectorclock.VectorClockRev(c_rev)
                if doc_vcr.is_newer(c_vcr):
                    continue
                if doc.same_content_as(Document(doc.doc_id, c_rev, c_doc)):
                    doc_vcr.maximize(c_vcr)
                    autoresolved = True
                    continue
                remaining_conflicts.append((c_rev, c_doc))
            if autoresolved:
                doc_vcr.increment(self._replica_uid)
                doc.rev = doc_vcr.as_str()
            self._replace_conflicts(doc, remaining_conflicts)

    def resolve_doc(self, doc, conflicted_doc_revs):
        cur_doc = self._get_doc(doc.doc_id)
        if cur_doc is None:
            cur_rev = None
        else:
            cur_rev = cur_doc.rev
        new_rev = self._ensure_maximal_rev(cur_rev, conflicted_doc_revs)
        superseded_revs = set(conflicted_doc_revs)
        remaining_conflicts = []
        cur_conflicts = self._conflicts[doc.doc_id]
        for c_rev, c_doc in cur_conflicts:
            if c_rev in superseded_revs:
                continue
            remaining_conflicts.append((c_rev, c_doc))
        doc.rev = new_rev
        if cur_rev in superseded_revs:
            self._put_and_update_indexes(cur_doc, doc)
        else:
            remaining_conflicts.append((new_rev, doc.get_json()))
        self._replace_conflicts(doc, remaining_conflicts)

    def delete_doc(self, doc):
        if doc.doc_id not in self._docs:
            raise errors.DocumentDoesNotExist
        if self._docs[doc.doc_id][1] in ('null', None):
            raise errors.DocumentAlreadyDeleted
        doc.make_tombstone()
        self.put_doc(doc)

    def create_index(self, index_name, *index_expressions):
        if index_name in self._indexes:
            if self._indexes[index_name]._definition == list(
                    index_expressions):
                return
            raise errors.IndexNameTakenError
        index = InMemoryIndex(index_name, list(index_expressions))
        for doc_id, (doc_rev, doc) in self._docs.iteritems():
            if doc is not None:
                index.add_json(doc_id, doc)
        self._indexes[index_name] = index

    def delete_index(self, index_name):
        try:
            del self._indexes[index_name]
        except KeyError:
            pass

    def list_indexes(self):
        definitions = []
        for idx in self._indexes.itervalues():
            definitions.append((idx._name, idx._definition))
        return definitions

    def get_from_index(self, index_name, *key_values):
        try:
            index = self._indexes[index_name]
        except KeyError:
            raise errors.IndexDoesNotExist
        doc_ids = index.lookup(key_values)
        result = []
        for doc_id in doc_ids:
            result.append(self._get_doc(doc_id, check_for_conflicts=True))
        return result

    def get_range_from_index(self, index_name, start_value=None,
                             end_value=None):
        """Return all documents with key values in the specified range."""
        try:
            index = self._indexes[index_name]
        except KeyError:
            raise errors.IndexDoesNotExist
        if isinstance(start_value, basestring):
            start_value = (start_value,)
        if isinstance(end_value, basestring):
            end_value = (end_value,)
        doc_ids = index.lookup_range(start_value, end_value)
        result = []
        for doc_id in doc_ids:
            result.append(self._get_doc(doc_id, check_for_conflicts=True))
        return result

    def get_index_keys(self, index_name):
        try:
            index = self._indexes[index_name]
        except KeyError:
            raise errors.IndexDoesNotExist
        keys = index.keys()
        # XXX inefficiency warning
        return list(set([tuple(key.split('\x01')) for key in keys]))

    def whats_changed(self, old_generation=0):
        changes = []
        relevant_tail = self._transaction_log[old_generation:]
        # We don't use len(self._transaction_log) because _transaction_log may
        # get mutated by a concurrent operation.
        cur_generation = old_generation + len(relevant_tail)
        last_trans_id = ''
        if relevant_tail:
            last_trans_id = relevant_tail[-1][1]
        elif self._transaction_log:
            last_trans_id = self._transaction_log[-1][1]
        seen = set()
        generation = cur_generation
        for doc_id, trans_id in reversed(relevant_tail):
            if doc_id not in seen:
                changes.append((doc_id, generation, trans_id))
                seen.add(doc_id)
            generation -= 1
        changes.reverse()
        return (cur_generation, last_trans_id, changes)

    def _force_doc_sync_conflict(self, doc):
        my_doc = self._get_doc(doc.doc_id)
        self._prune_conflicts(doc, vectorclock.VectorClockRev(doc.rev))
        self._conflicts.setdefault(doc.doc_id, []).append(
            (my_doc.rev, my_doc.get_json()))
        doc.has_conflicts = True
        self._put_and_update_indexes(my_doc, doc)


class InMemoryIndex(object):
    """Interface for managing an Index."""

    def __init__(self, index_name, index_definition):
        self._name = index_name
        self._definition = index_definition
        self._values = {}
        parser = query_parser.Parser()
        self._getters = parser.parse_all(self._definition)

    def evaluate_json(self, doc):
        """Determine the 'key' after applying this index to the doc."""
        raw = json.loads(doc)
        return self.evaluate(raw)

    def evaluate(self, obj):
        """Evaluate a dict object, applying this definition."""
        all_rows = [[]]
        for getter in self._getters:
            new_rows = []
            keys = getter.get(obj)
            if not keys:
                return []
            for key in keys:
                new_rows.extend([row + [key] for row in all_rows])
            all_rows = new_rows
        all_rows = ['\x01'.join(row) for row in all_rows]
        return all_rows

    def add_json(self, doc_id, doc):
        """Add this json doc to the index."""
        keys = self.evaluate_json(doc)
        if not keys:
            return
        for key in keys:
            self._values.setdefault(key, []).append(doc_id)

    def remove_json(self, doc_id, doc):
        """Remove this json doc from the index."""
        keys = self.evaluate_json(doc)
        if keys:
            for key in keys:
                doc_ids = self._values[key]
                doc_ids.remove(doc_id)
                if not doc_ids:
                    del self._values[key]

    def _find_non_wildcards(self, values):
        """Check if this should be a wildcard match.

        Further, this will raise an exception if the syntax is improperly
        defined.

        :return: The offset of the last value we need to match against.
        """
        if len(values) != len(self._definition):
            raise errors.InvalidValueForIndex()
        is_wildcard = False
        last = 0
        for idx, val in enumerate(values):
            if val.endswith('*'):
                if val != '*':
                    # We have an 'x*' style wildcard
                    if is_wildcard:
                        # We were already in wildcard mode, so this is invalid
                        raise errors.InvalidGlobbing
                    last = idx + 1
                is_wildcard = True
            else:
                if is_wildcard:
                    # We were in wildcard mode, we can't follow that with
                    # non-wildcard
                    raise errors.InvalidGlobbing
                last = idx + 1
        if not is_wildcard:
            return -1
        return last

    def lookup(self, values):
        """Find docs that match the values."""
        last = self._find_non_wildcards(values)
        if last == -1:
            return self._lookup_exact(values)
        else:
            return self._lookup_prefix(values[:last])

    def lookup_range(self, start_values, end_values):
        """Find docs within the range."""
        # TODO: Wildly inefficient, which is unlikely to be a problem for the
        # inmemory implementation.
        if start_values:
            self._find_non_wildcards(start_values)
            start_values = get_prefix(start_values)
        if end_values:
            if self._find_non_wildcards(end_values) == -1:
                exact = True
            else:
                exact = False
            end_values = get_prefix(end_values)
        found = []
        for key, doc_ids in sorted(self._values.iteritems()):
            if start_values and start_values > key:
                continue
            if end_values and end_values < key:
                if exact:
                    break
                else:
                    if not key.startswith(end_values):
                        break
            found.extend(doc_ids)
        return found

    def keys(self):
        """Find the indexed keys."""
        return self._values.keys()

    def _lookup_prefix(self, value):
        """Find docs that match the prefix string in values."""
        # TODO: We need a different data structure to make prefix style fast,
        #       some sort of sorted list would work, but a plain dict doesn't.
        key_prefix = get_prefix(value)
        all_doc_ids = []
        for key, doc_ids in sorted(self._values.iteritems()):
            if key.startswith(key_prefix):
                all_doc_ids.extend(doc_ids)
        return all_doc_ids

    def _lookup_exact(self, value):
        """Find docs that match exactly."""
        key = '\x01'.join(value)
        if key in self._values:
            return self._values[key]
        return ()


class InMemorySyncTarget(CommonSyncTarget):

    def get_sync_info(self, source_replica_uid):
        source_gen, source_trans_id = self._db._get_replica_gen_and_trans_id(
            source_replica_uid)
        my_gen, my_trans_id = self._db._get_generation_info()
        return (
            self._db._replica_uid, my_gen, my_trans_id, source_gen,
            source_trans_id)

    def record_sync_info(self, source_replica_uid, source_replica_generation,
                         source_transaction_id):
        if self._trace_hook:
            self._trace_hook('record_sync_info')
        self._db._set_replica_gen_and_trans_id(
            source_replica_uid, source_replica_generation,
            source_transaction_id)



# Copyright 2012 Canonical Ltd.
#
# This file is part of u1db.
#
# u1db is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# u1db is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with u1db.  If not, see <http://www.gnu.org/licenses/>.
"""U1DB Basic Auth authorisation WSGI middleware."""
import httplib
import json

from wsgiref.util import shift_path_info


class Unauthorized(Exception):
    """User authorization failed."""


class BasicAuthMiddleware(object):
    """U1DB Basic Auth Authorisation WSGI middleware."""

    def __init__(self, app, prefix):
        self.app = app
        self.prefix = prefix

    def _error(self, start_response, status, description, message=None):
        start_response("%d %s" % (status, httplib.responses[status]),
                       [('content-type', 'application/json')])
        err = {"error": description}
        if message:
            err['message'] = message
        return [json.dumps(err)]

    def __call__(self, environ, start_response):
        if self.prefix and not environ['PATH_INFO'].startswith(self.prefix):
            return self._error(start_response, 400, "bad request")
        auth = environ.get('HTTP_AUTHORIZATION')
        if not auth:
            return self._error(start_response, 401, "unauthorized",
                               "Missing Basic Authentication.")
        scheme, encoded = auth.split(None, 1)
        if scheme.lower() != 'basic':
            return self._error(
                start_response, 401, "unauthorized",
                "Missing Basic Authentication")
        user, password = encoded.decode('base64').split(':', 1)
        try:
            self.verify_user(environ, user, password)
        except Unauthorized:
            return self._error(
                start_response, 401, "unauthorized",
                "Incorrect password or login.")
        del environ['HTTP_AUTHORIZATION']
        shift_path_info(environ)
        return self.app(environ, start_response)

    def verify_user(self, environ, username, password):
        raise NotImplementedError(self.verify_user)



# -*- coding: utf-8 -*-
# errors.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""
Soledad errors.
"""

from .l2db import errors
from .l2db.remote import http_errors


def register_exception(cls):
    """
    A small decorator that registers exceptions in u1db maps.
    """
    # update u1db "wire description to status" and "wire description to
    # exception" maps.
    http_errors.wire_description_to_status.update({
        cls.wire_description: cls.status})
    errors.wire_description_to_exc.update({
        cls.wire_description: cls})
    # do not modify the exception
    return cls


class SoledadError(errors.U1DBError):

    """
    Base Soledad HTTP errors.
    """
    pass


#
# Authorization errors
#


class DatabaseAccessError(Exception):
    pass


@register_exception
class InvalidAuthTokenError(errors.Unauthorized):

    """
    Exception raised when failing to get authorization for some action because
    the provided token either does not exist in the tokens database, has a
    distinct structure from the expected one, or is associated with a user
    with a distinct uuid than the one provided by the client.
    """

    wire_descrition = "invalid auth token"
    status = 401


#
# SoledadBackend errors
# u1db error statuses also have to be updated
http_errors.ERROR_STATUSES = set(
    http_errors.wire_description_to_status.values())


class InvalidURLError(Exception):
    """
    Exception raised when Soledad encounters a malformed URL.
    """


class BackendNotReadyError(SoledadError):
    """
    Generic exception raised when the backend is not ready to dispatch a client
    request.
    """
    wire_description = "backend not ready"
    status = 500


class WrongCouchSchemaVersionError(SoledadError):
    """
    Raised in case there is a user database with wrong couch schema version.
    """


class MissingCouchConfigDocumentError(SoledadError):
    """
    Raised if a database has documents but lacks the couch config document.
    """



# -*- coding: utf-8 -*-
# __init__.py
# Copyright (C) 2015 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""A U1DB backend that uses CouchDB as its persistence layer."""


import json
import re
import uuid
import binascii


from StringIO import StringIO
from urlparse import urljoin
from contextlib import contextmanager


from couchdb.client import Server, Database
from couchdb.http import (
    ResourceConflict,
    ResourceNotFound,
    Session,
    urljoin as couch_urljoin,
    Resource,
)
from leap.soledad.common.l2db.errors import (
    DatabaseDoesNotExist,
    InvalidGeneration,
    RevisionConflict,
)
from leap.soledad.common.l2db.remote import http_app


from .support import MultipartWriter
from leap.soledad.common.errors import InvalidURLError
from leap.soledad.common.document import ServerDocument
from leap.soledad.common.backend import SoledadBackend


COUCH_TIMEOUT = 120  # timeout for transfers between Soledad server and Couch


def list_users_dbs(couch_url):
    """
    Retrieves a list with all databases that starts with 'user-' on CouchDB.
    Those databases belongs to users. So, the list will contain all the
    database names in the form of 'user-{uuid4}'.

    :param couch_url: The couch url with needed credentials
    :type couch_url: str

    :return: The list of all database names from users.
    :rtype: [str]
    """
    with couch_server(couch_url) as server:
        users = [dbname for dbname in server if dbname.startswith('user-')]
    return users


# monkey-patch the u1db http app to use ServerDocument
http_app.Document = ServerDocument


@contextmanager
def couch_server(url):
    """
    Provide a connection to a couch server and cleanup after use.

    For database creation and deletion we use an ephemeral connection to the
    couch server. That connection has to be properly closed, so we provide it
    as a context manager.

    :param url: The URL of the Couch server.
    :type url: str
    """
    session = Session(timeout=COUCH_TIMEOUT)
    server = Server(url=url, full_commit=False, session=session)
    yield server


def _get_gen_doc_id(gen):
    return 'gen-%s' % str(gen).zfill(10)


GENERATION_KEY = 'gen'
TRANSACTION_ID_KEY = 'trans_id'
REPLICA_UID_KEY = 'replica_uid'
DOC_ID_KEY = 'doc_id'
SCHEMA_VERSION_KEY = 'schema_version'

CONFIG_DOC_ID = '_local/config'
SYNC_DOC_ID_PREFIX = '_local/sync_'
SCHEMA_VERSION = 1


class CouchDatabase(object):
    """
    Holds CouchDB related code.
    This class gives methods to encapsulate database operations and hide
    CouchDB details from backend code.
    """

    @classmethod
    def open_database(cls, url, create, replica_uid=None,
                      database_security=None):
        """
        Open a U1DB database using CouchDB as backend.

        :param url: the url of the database replica
        :type url: str
        :param create: should the replica be created if it does not exist?
        :type create: bool
        :param replica_uid: an optional unique replica identifier
        :type replica_uid: str
        :param database_security: security rules as CouchDB security doc
        :type database_security: dict

        :return: the database instance
        :rtype: SoledadBackend

        :raise DatabaseDoesNotExist: Raised if database does not exist.
        """
        # get database from url
        m = re.match('(^https?://[^/]+)/(.+)$', url)
        if not m:
            raise InvalidURLError
        url = m.group(1)
        dbname = m.group(2)
        with couch_server(url) as server:
            if dbname not in server:
                if create:
                    server.create(dbname)
                else:
                    raise DatabaseDoesNotExist()
        db = cls(url, dbname, ensure_security=create,
                 database_security=database_security)
        return SoledadBackend(
            db, replica_uid=replica_uid)

    def __init__(self, url, dbname, ensure_security=False,
                 database_security=None):
        """
        :param url: Couch server URL with necessary credentials
        :type url: string
        :param dbname: Couch database name
        :type dbname: string
        :param ensure_security: will PUT a _security ddoc if set
        :type ensure_security: bool
        :param database_security: security rules as CouchDB security doc
        :type database_security: dict
        """
        self._session = Session(timeout=COUCH_TIMEOUT)
        self._url = url
        self._dbname = dbname
        self._database = self.get_couch_database(url, dbname)
        self.batching = False
        self.batch_generation = None
        self.batch_docs = {}
        if ensure_security:
            self.ensure_security_ddoc(database_security)

    def batch_start(self):
        self.batching = True
        self.batch_generation = self.get_generation_info()
        ids = set(row.id for row in self._database.view('_all_docs'))
        self.batched_ids = ids

    def batch_end(self):
        self.batching = False
        self.batch_generation = None
        self.__perform_batch()

    def get_couch_database(self, url, dbname):
        """
        Generate a couchdb.Database instance given a url and dbname.

        :param url: CouchDB's server url with credentials
        :type url: str
        :param dbname: Database name
        :type dbname: str

        :return: couch library database instance
        :rtype: couchdb.Database

        :raise DatabaseDoesNotExist: Raised if database does not exist.
        """
        try:
            return Database(
                urljoin(url, dbname),
                self._session)
        except ResourceNotFound:
            raise DatabaseDoesNotExist()

    def ensure_security_ddoc(self, security_config=None):
        """
        Make sure that only soledad user is able to access this database as
        an unprivileged member, meaning that administration access will
        be forbidden even inside an user database.
        The goal is to make sure that only the lowest access level is given
        to the unprivileged CouchDB user set on the server process.
        This is achieved by creating a _security design document, see:
        http://docs.couchdb.org/en/latest/api/database/security.html

        :param security_config: security configuration parsed from conf file
        :type security_config: dict
        """
        security_config = security_config or {}
        security = self._database.resource.get_json('_security')[2]
        security['members'] = {'names': [], 'roles': []}
        security['members']['names'] = security_config.get('members',
                                                           ['soledad'])
        security['members']['roles'] = security_config.get('members_roles', [])
        security['admins'] = {'names': [], 'roles': []}
        security['admins']['names'] = security_config.get('admins', [])
        security['admins']['roles'] = security_config.get('admins_roles', [])
        self._database.resource.put_json('_security', body=security)

    def delete_database(self):
        """
        Delete a U1DB CouchDB database.
        """
        with couch_server(self._url) as server:
            del(server[self._dbname])

    def set_replica_uid(self, replica_uid):
        """
        Force the replica uid to be set.

        :param replica_uid: The new replica uid.
        :type replica_uid: str
        """
        try:
            # set on existent config document
            doc = self._database[CONFIG_DOC_ID]
            doc[REPLICA_UID_KEY] = replica_uid
        except ResourceNotFound:
            # or create the config document
            doc = {
                '_id': CONFIG_DOC_ID,
                REPLICA_UID_KEY: replica_uid,
                SCHEMA_VERSION_KEY: SCHEMA_VERSION,
            }
        self._database.save(doc)

    def get_replica_uid(self):
        """
        Get the replica uid.

        :return: The replica uid.
        :rtype: str
        """
        try:
            # grab replica_uid from server
            doc = self._database[CONFIG_DOC_ID]
            replica_uid = doc[REPLICA_UID_KEY]
            return replica_uid
        except ResourceNotFound:
            # create a unique replica_uid
            replica_uid = uuid.uuid4().hex
            self.set_replica_uid(replica_uid)
            return replica_uid

    def close(self):
        self._database = None

    def get_all_docs(self, include_deleted=False):
        """
        Get the JSON content for all documents in the database.

        :param include_deleted: If set to True, deleted documents will be
                                returned with empty content. Otherwise deleted
                                documents will not be included in the results.
        :type include_deleted: bool

        :return: (generation, [ServerDocument])
            The current generation of the database, followed by a list of all
            the documents in the database.
        :rtype: (int, [ServerDocument])
        """

        generation, _ = self.get_generation_info()
        results = list(
            self._get_docs(None, True, include_deleted))
        return (generation, results)

    def get_docs(self, doc_ids, check_for_conflicts=True,
                 include_deleted=False):
        """
        Get the JSON content for many documents.

        :param doc_ids: A list of document identifiers or None for all.
        :type doc_ids: list
        :param check_for_conflicts: If set to False, then the conflict check
                                    will be skipped, and 'None' will be
                                    returned instead of True/False.
        :type check_for_conflicts: bool
        :param include_deleted: If set to True, deleted documents will be
                                returned with empty content. Otherwise deleted
                                documents will not be included in the results.
        :return: iterable giving the Document object for each document id
                 in matching doc_ids order.
        :rtype: iterable
        """
        return self._get_docs(doc_ids, check_for_conflicts, include_deleted)

    def _get_docs(self, doc_ids, check_for_conflicts, include_deleted):
        """
        Use couch's `_all_docs` view to get the documents indicated in
        `doc_ids`,

        :param doc_ids: A list of document identifiers or None for all.
        :type doc_ids: list
        :param check_for_conflicts: If set to False, then the conflict check
                                    will be skipped, and 'None' will be
                                    returned instead of True/False.
        :type check_for_conflicts: bool
        :param include_deleted: If set to True, deleted documents will be
                                returned with empty content. Otherwise deleted
                                documents will not be included in the results.

        :return: iterable giving the Document object for each document id
                 in matching doc_ids order.
        :rtype: iterable
        """
        params = {'include_docs': 'true', 'attachments': 'true'}
        if doc_ids is not None:
            params['keys'] = doc_ids
        view = self._database.view("_all_docs", **params)
        for row in view.rows:
            result = row['doc']
            doc = self.__parse_doc_from_couch(
                result, result['_id'], check_for_conflicts=check_for_conflicts)
            # filter out non-u1db or deleted documents
            if not doc or (not include_deleted and doc.is_tombstone()):
                continue
            yield doc

    def get_doc(self, doc_id, check_for_conflicts=False):
        """
        Extract the document from storage.

        This can return None if the document doesn't exist.

        :param doc_id: The unique document identifier
        :type doc_id: str
        :param check_for_conflicts: If set to False, then the conflict check
                                    will be skipped.
        :type check_for_conflicts: bool

        :return: The document.
        :rtype: ServerDocument
        """
        doc_from_batch = self.__check_batch_before_get(doc_id)
        if doc_from_batch:
            return doc_from_batch
        if self.batching and doc_id not in self.batched_ids:
            return None
        if doc_id not in self._database:
            return None
        # get document with all attachments (u1db content and eventual
        # conflicts)
        result = self.json_from_resource([doc_id], attachments=True)
        return self.__parse_doc_from_couch(result, doc_id, check_for_conflicts)

    def __check_batch_before_get(self, doc_id):
        """
        If doc_id is staged for batching, then we need to commit the batch
        before going ahead. This avoids consistency problems, like trying to
        get a document that isn't persisted and processing like it is missing.

        :param doc_id: The unique document identifier
        :type doc_id: str
        """
        if doc_id in self.batch_docs:
            couch_doc = self.batch_docs[doc_id]
            rev = self.__perform_batch(doc_id)
            couch_doc['_rev'] = rev
            self.batched_ids.add(doc_id)
            return self.__parse_doc_from_couch(couch_doc, doc_id, True)
        return None

    def __perform_batch(self, doc_id=None):
        status = self._database.update(self.batch_docs.values())
        rev = None
        for ok, stored_doc_id, rev_or_error in status:
            if not ok:
                error = rev_or_error
                if type(error) is ResourceConflict:
                    raise RevisionConflict
                raise error
            elif doc_id == stored_doc_id:
                rev = rev_or_error
        self.batch_docs.clear()
        return rev

    def __parse_doc_from_couch(self, result, doc_id,
                               check_for_conflicts=False):
        # restrict to u1db documents
        if 'u1db_rev' not in result:
            return None
        doc = ServerDocument(doc_id, result['u1db_rev'])
        # set contents or make tombstone
        if '_attachments' not in result \
                or 'u1db_content' not in result['_attachments']:
            doc.make_tombstone()
        else:
            doc.content = json.loads(
                binascii.a2b_base64(
                    result['_attachments']['u1db_content']['data']))
        # determine if there are conflicts
        if check_for_conflicts \
                and '_attachments' in result \
                and 'u1db_conflicts' in result['_attachments']:
            doc.set_conflicts(
                self._build_conflicts(
                    doc.doc_id,
                    json.loads(binascii.a2b_base64(
                        result['_attachments']['u1db_conflicts']['data']))))
        # store couch revision
        doc.couch_rev = result['_rev']
        return doc

    def _build_conflicts(self, doc_id, attached_conflicts):
        """
        Build the conflicted documents list from the conflicts attachment
        fetched from a couch document.

        :param attached_conflicts: The document's conflicts as fetched from a
                                   couch document attachment.
        :type attached_conflicts: dict
        """
        conflicts = []
        for doc_rev, content in attached_conflicts:
            doc = ServerDocument(doc_id, doc_rev)
            if content is None:
                doc.make_tombstone()
            else:
                doc.content = content
            conflicts.append(doc)
        return conflicts

    def get_trans_id_for_gen(self, generation):
        """
        Get the transaction id corresponding to a particular generation.

        :param generation: The generation for which to get the transaction id.
        :type generation: int

        :return: The transaction id for C{generation}.
        :rtype: str

        :raise InvalidGeneration: Raised when the generation does not exist.
        """
        if generation == 0:
            return ''
        log = self._get_transaction_log(start=generation, end=generation)
        if not log:
            raise InvalidGeneration
        _, _, trans_id = log[0]
        return trans_id

    def get_replica_gen_and_trans_id(self, other_replica_uid):
        """
        Return the last known generation and transaction id for the other db
        replica.

        When you do a synchronization with another replica, the Database keeps
        track of what generation the other database replica was at, and what
        the associated transaction id was.  This is used to determine what data
        needs to be sent, and if two databases are claiming to be the same
        replica.

        :param other_replica_uid: The identifier for the other replica.
        :type other_replica_uid: str

        :return: A tuple containing the generation and transaction id we
                 encountered during synchronization. If we've never
                 synchronized with the replica, this is (0, '').
        :rtype: (int, str)
        """
        doc_id = '%s%s' % (SYNC_DOC_ID_PREFIX, other_replica_uid)
        try:
            doc = self._database[doc_id]
        except ResourceNotFound:
            doc = {
                '_id': doc_id,
                GENERATION_KEY: 0,
                REPLICA_UID_KEY: str(other_replica_uid),
                TRANSACTION_ID_KEY: '',
            }
            self._database.save(doc)
        gen, trans_id = doc[GENERATION_KEY], doc[TRANSACTION_ID_KEY]
        return gen, trans_id

    def get_doc_conflicts(self, doc_id, couch_rev=None):
        """
        Get the conflicted versions of a document.

        If the C{couch_rev} parameter is not None, conflicts for a specific
        document's couch revision are returned.

        :param couch_rev: The couch document revision.
        :type couch_rev: str

        :return: A list of conflicted versions of the document.
        :rtype: list
        """
        # request conflicts attachment from server
        params = {}
        conflicts = []
        if couch_rev is not None:
            params['rev'] = couch_rev  # restric document's couch revision
        else:
            # TODO: move into resource logic!
            first_entry = self.get_doc(doc_id, check_for_conflicts=True)
            conflicts.append(first_entry)

        try:
            response = self.json_from_resource([doc_id, 'u1db_conflicts'],
                                               **params)
            return conflicts + self._build_conflicts(
                doc_id, json.loads(response.read()))
        except ResourceNotFound:
            return []

    def set_replica_gen_and_trans_id(
            self, other_replica_uid, other_generation, other_transaction_id):
        """
        Set the last-known generation and transaction id for the other
        database replica.

        We have just performed some synchronization, and we want to track what
        generation the other replica was at. See also
        get_replica_gen_and_trans_id.

        :param other_replica_uid: The U1DB identifier for the other replica.
        :type other_replica_uid: str
        :param other_generation: The generation number for the other replica.
        :type other_generation: int
        :param other_transaction_id: The transaction id associated with the
                                     generation.
        :type other_transaction_id: str
        """
        doc_id = '%s%s' % (SYNC_DOC_ID_PREFIX, other_replica_uid)
        try:
            doc = self._database[doc_id]
        except ResourceNotFound:
            doc = {'_id': doc_id}
        doc[GENERATION_KEY] = other_generation
        doc[TRANSACTION_ID_KEY] = other_transaction_id
        self._database.save(doc)

    def get_transaction_log(self):
        """
        This is only for the test suite, it is not part of the api.

        :return: The complete transaction log.
        :rtype: [(str, str)]
        """
        log = self._get_transaction_log()
        return map(lambda i: (i[1], i[2]), log)

    def _get_gen_docs(
            self, start=0, end=9999999999, descending=None, limit=None):
        params = {}
        if descending:
            params['descending'] = 'true'
            # honor couch way of traversing the view tree in reverse order
            start, end = end, start
        params['startkey'] = _get_gen_doc_id(start)
        params['endkey'] = _get_gen_doc_id(end)
        params['include_docs'] = 'true'
        if limit:
            params['limit'] = limit
        view = self._database.view("_all_docs", **params)
        return view.rows

    def _get_transaction_log(self, start=0, end=9999999999):
        # get current gen and trans_id
        rows = self._get_gen_docs(start=start, end=end)
        log = []
        for row in rows:
            doc = row['doc']
            log.append((
                doc[GENERATION_KEY],
                doc[DOC_ID_KEY],
                doc[TRANSACTION_ID_KEY]))
        return log

    def whats_changed(self, old_generation=0):
        """
        Return a list of documents that have changed since old_generation.

        :param old_generation: The generation of the database in the old
                               state.
        :type old_generation: int

        :return: (generation, trans_id, [(doc_id, generation, trans_id),...])
                 The current generation of the database, its associated
                 transaction id, and a list of of changed documents since
                 old_generation, represented by tuples with for each document
                 its doc_id and the generation and transaction id corresponding
                 to the last intervening change and sorted by generation (old
                 changes first)
        :rtype: (int, str, [(str, int, str)])
        """
        changes = []
        cur_generation, last_trans_id = self.get_generation_info()
        relevant_tail = self._get_transaction_log(start=old_generation + 1)
        seen = set()
        for generation, doc_id, trans_id in reversed(relevant_tail):
            if doc_id not in seen:
                changes.append((doc_id, generation, trans_id))
                seen.add(doc_id)
        changes.reverse()
        return (cur_generation, last_trans_id, changes)

    def get_generation_info(self):
        """
        Return the current generation.

        :return: A tuple containing the current generation and transaction id.
        :rtype: (int, str)
        """
        if self.batching and self.batch_generation:
            return self.batch_generation
        rows = self._get_gen_docs(descending=True, limit=1)
        if not rows:
            return 0, ''
        gen_doc = rows.pop()['doc']
        return gen_doc[GENERATION_KEY], gen_doc[TRANSACTION_ID_KEY]

    def json_from_resource(self, doc_path, **kwargs):
        """
        Get a resource from it's path and gets a doc's JSON using provided
        parameters.

        :param doc_path: The path to resource.
        :type doc_path: [str]

        :return: The request's data parsed from JSON to a dict.
        :rtype: dict
        """
        if doc_path is not None:
            resource = self._database.resource(*doc_path)
        else:
            resource = self._database.resource()
        _, _, data = resource.get_json(**kwargs)
        return data

    def _allocate_new_generation(self, doc_id, transaction_id):
        """
        Allocate a new generation number for a document modification.

        We need to allocate a new generation to this document modification by
        creating a new gen doc. In order to avoid concurrent database updates
        from allocating the same new generation, we will try to create the
        document until we succeed, meaning that no other piece of code holds
        the same generation number as ours.

        The loop below would only be executed more than once if:

          1. there's more than one thread trying to modify the user's database,
             and

          2. the execution of getting the current generation and saving the gen
             doc different threads get interleaved (one of them will succeed
             and the others will fail and try again).

        Number 1 only happens when more than one user device is syncing at the
        same time. Number 2 depends on not-so-frequent coincidence of
        code execution.

        Also, in the race between threads for a generation number there's
        always one thread that wins. so if there are N threads in the race, the
        expected number of repetitions of the loop for each thread would be
        N/2. If N is equal to the number of devices that the user has, the
        number of possible repetitions of the loop should always be low.
        """
        while True:
            try:
                # add the gen document
                gen, _ = self.get_generation_info()
                new_gen = gen + 1
                gen_doc = {
                    '_id': _get_gen_doc_id(new_gen),
                    GENERATION_KEY: new_gen,
                    DOC_ID_KEY: doc_id,
                    TRANSACTION_ID_KEY: transaction_id,
                }
                self._database.save(gen_doc)
                break  # succeeded allocating a new generation, proceed
            except ResourceConflict:
                pass  # try again!

    def save_document(self, old_doc, doc, transaction_id):
        """
        Put the document in the Couch backend database.

        Note that C{old_doc} must have been fetched with the parameter
        C{check_for_conflicts} equal to True, so we can properly update the
        new document using the conflict information from the old one.

        :param old_doc: The old document version.
        :type old_doc: ServerDocument
        :param doc: The document to be put.
        :type doc: ServerDocument

        :raise RevisionConflict: Raised when trying to update a document but
                                 couch revisions mismatch.
        """
        attachments = {}  # we save content and conflicts as attachments
        parts = []  # and we put it using couch's multipart PUT
        # save content as attachment
        if doc.is_tombstone() is False:
            content = doc.get_json()
            attachments['u1db_content'] = {
                'follows': True,
                'content_type': 'application/octet-stream',
                'length': len(content),
            }
            parts.append(content)

        # save conflicts as attachment
        if doc.has_conflicts is True:
            conflicts = json.dumps(
                map(lambda cdoc: (cdoc.rev, cdoc.content),
                    doc.get_conflicts()))
            attachments['u1db_conflicts'] = {
                'follows': True,
                'content_type': 'application/octet-stream',
                'length': len(conflicts),
            }
            parts.append(conflicts)

        # build the couch document
        couch_doc = {
            '_id': doc.doc_id,
            'u1db_rev': doc.rev,
            '_attachments': attachments,
        }
        # if we are updating a doc we have to add the couch doc revision
        if old_doc is not None and hasattr(old_doc, 'couch_rev'):
            couch_doc['_rev'] = old_doc.couch_rev
        # prepare the multipart PUT
        if not self.batching:
            buf = StringIO()
            envelope = MultipartWriter(buf)
            # the order in which attachments are described inside the
            # serialization of the couch document must match the order in
            # which they are actually written in the multipart structure.
            # Because of that, we use `sorted_keys=True` in the json
            # serialization (so "u1db_conflicts" comes before
            # "u1db_content" on the couch document attachments
            # description), and also reverse the order of the parts before
            # writing them, so the "conflict" part is written before the
            # "content" part.
            envelope.add(
                'application/json',
                json.dumps(couch_doc, sort_keys=True))
            parts.reverse()
            for part in parts:
                envelope.add('application/octet-stream', part)
            envelope.close()
            # try to save and fail if there's a revision conflict
            try:
                resource = self._new_resource()
                resource.put_json(
                    doc.doc_id, body=str(buf.getvalue()),
                    headers=envelope.headers)
            except ResourceConflict:
                raise RevisionConflict()
        else:
            for name, attachment in attachments.items():
                del attachment['follows']
                del attachment['length']
                index = 0 if name is 'u1db_content' else 1
                attachment['data'] = binascii.b2a_base64(
                    parts[index]).strip()
            couch_doc['_attachments'] = attachments
            self.batch_docs[doc.doc_id] = couch_doc
            last_gen, last_trans_id = self.batch_generation
            self.batch_generation = (last_gen + 1, transaction_id)

        self._allocate_new_generation(doc.doc_id, transaction_id)

    def _new_resource(self, *path):
        """
        Return a new resource for accessing a couch database.

        :return: A resource for accessing a couch database.
        :rtype: couchdb.http.Resource
        """
        # Workaround for: https://leap.se/code/issues/5448
        url = couch_urljoin(self._database.resource.url, *path)
        resource = Resource(url, Session(timeout=COUCH_TIMEOUT))
        resource.credentials = self._database.resource.credentials
        resource.headers = self._database.resource.headers.copy()
        return resource
